# version 1.2.0
import functions_framework
from google.cloud import storage
import os
import subprocess
import tempfile
from pathlib import Path
import json
import re

# Initialize Google Cloud Storage client
storage_client = storage.Client()


@functions_framework.cloud_event
def process_mdf_file(cloud_event):
    print("------------------------\nReceived new cloud_event: ", cloud_event)
    data = cloud_event.data

    file_name = data["name"]

    print("file_name: ", file_name)
    # check if file_name indicates a folder creation event
    if file_name[-1] == "/" or data["size"] == "0":
        print(f"{file_name} is not a file creation event - exiting")
        return

    # check if file_name is valid MDF
    file_name_suffix = file_name.split(".")[-1]
    if file_name_suffix not in ["MF4", "MFC", "MFE", "MFM"]:
        print(f"{file_name} is not a valid MDF file - exiting")
        return

    bucket_input = data["bucket"]
    bucket_output = (
        bucket_input + "-parquet"
    )  # assumes output bucket is named correctly

    print(f"Processing file: {file_name} in bucket: {bucket_input}")

    files_uploaded = 0
    decoder = "mdf2parquet_decode"
    dbc_groups = True

    with tempfile.TemporaryDirectory() as temp:
        tmp_dir = Path(temp)
        tmp_input_dir = tmp_dir / "input"
        tmp_output_dir = tmp_dir / "output"
        tmp_input_dir.mkdir()
        tmp_output_dir.mkdir()

        bucket = storage_client.bucket(bucket_input)
        blob = bucket.blob(file_name)
        fs_logfile_path = tmp_input_dir / file_name

        # Create directories if they do not exist
        if not fs_logfile_path.parent.exists():
            os.makedirs(fs_logfile_path.parent)

        # Download the MF4 file
        blob.download_to_filename(fs_logfile_path)
        print(f"Downloaded MDF log file to {fs_logfile_path}")

        parts = file_name.split("/")
        device_id = ""
        # Check if the path has at least 3 parts
        if len(parts) == 3 and re.match("[0-9A-F]{8}$", parts[0]):
            # The device_id is expected to be the first part of the path
            device_id = parts[0]
            print(f"Device ID: {device_id}")
        else:
            print(
                f"WARNING: Unable to extract device_id - log file does not use CANedge S3 path"
            )

        # Handling dbc-groups.json
        device_specific_dbc_files = []
        if dbc_groups and device_id != "":
            try:
                dbc_groups_blob = bucket.blob("dbc-groups.json")
                dbc_groups_path = tmp_input_dir / "dbc-groups.json"
                dbc_groups_blob.download_to_filename(dbc_groups_path)
                with open(dbc_groups_path, "r") as file:
                    groups_data = json.load(file)
                for group in groups_data["dbc_groups"]:
                    if device_id in group["devices"]:
                        device_specific_dbc_files = group["dbc_files"]
                        break
            except Exception as e:
                print(f"Note: Applying all DBC files across all devices")

        # Download DBC files
        dbc_files = []
        for type_prefix in ["can", "lin"]:
            for blob in bucket.list_blobs(prefix=type_prefix):
                if blob.name.endswith(".dbc") and (
                    not device_specific_dbc_files
                    or blob.name in device_specific_dbc_files
                ):
                    local_dbc_path = tmp_input_dir / blob.name
                    blob.download_to_filename(local_dbc_path)
                    dbc_files.append(blob.name)
                    print(f"Downloaded {blob.name} to {local_dbc_path}")

        # Check if DBC files were found
        if len(dbc_files) == 0:
            print("No DBC files found. Exiting function.")
            return

        # Download the passwords.json file if needed
        if fs_logfile_path.suffix in [".MFE", ".MFM"]:
            fs_passwords_file_path = tmp_input_dir / "passwords.json"
            try:
                dbc_groups_blob = bucket.blob("passwords.json")
                dbc_groups_blob.download_to_filename(fs_passwords_file_path)
                print("Downloaded passwords.json successfully!")
            except Exception as e:
                print(
                    f"ERROR: The {fs_logfile_path.suffix} is encrypted, but no passwords.json file found in {bucket_input}/ - exiting"
                )
                return

        # Execute the decoder
        subprocess.run(["cp", "./" + decoder, tmp_input_dir])
        subprocess.run(
            [os.path.join(tmp_input_dir, decoder), "-v"],
            cwd=str(tmp_input_dir),
        )
        subprocess.run(
            [
                os.path.join(tmp_input_dir, decoder),
                "-i",
                str(fs_logfile_path),
                "-O",
                str(tmp_output_dir),
                "--verbosity=5",
                "-X",
            ],
            cwd=tmp_input_dir,
        )

        # Upload output files
        output_bucket = storage_client.bucket(bucket_output)
        for output_file in tmp_output_dir.rglob("*"):
            if output_file.is_file():
                output_blob = output_bucket.blob(
                    str(output_file.relative_to(tmp_output_dir))
                )
                output_blob.upload_from_filename(output_file)
                print(
                    f"Uploaded {output_file} as {output_file.name} to {output_bucket.name}"
                )
                files_uploaded += 1

        if files_uploaded > 0:
            print(f"{files_uploaded} parquet files were uploaded successfully.")
        else:
            print("No parquet files were uploaded.")
